import numpy

from . import util
from .train import Trainer
from .exceptions import ShapeError


class Model(object):
    '''Model base class.'''
    name = 'model'
    device = 'cpu'
    Trainer = Trainer
    ops = None
    output_shape = None
    input_shape = None
    
    def __init__(self, *args, **kwargs):
        self.layers = []
        kwargs = self.update_defaults(*args, **kwargs)
        if self.ops is None:
            self.ops = util.get_ops(self.device)
        self.setup(*args, **kwargs)

    def update_defaults(self, *args, **kwargs):
        new_kwargs = {}
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)
            else:
                new_kwargs[key] = value
        return new_kwargs
    
    def setup(self, *args, **kwargs):
        pass

    def check_shape(self, x, is_batch):
        if is_batch:
            if len(x.shape) != 2:
                raise ShapeError.expected_batch(locals(), globals())
            if x.shape[1] != self.nr_in:
                dims = (x.shape[1], self.nr_in)
                raise ShapeError.dim_mismatch(dims, locals(), globals())
        else:
            if x.shape[0] != self.nr_in:
                dims = (x.shape[0], self.nr_in)
                raise ShapeError.dim_mismatch(dims, locals(), globals())

    def __call__(self, x):
        '''Predict a single x.'''
        if not self.is_initialized:
            self.set_weights(initialize=True, example=x)
        is_batch = self.is_batch(x)
        self.check_shape(x, is_batch)
        if is_batch:
            return self.predict_batch(x)
        else:
            return self.predict_one(x)

    def is_batch(self, X):
        if hasattr(X, 'shape') and len(X.shape) >= 2:
            return True
        else:
            return False
    
    def predict_one(self, x):
        X = self.ops.expand_dims(x, axis=0)
        return self.predict_batch(X)[0]

    def predict_batch(self, X):
        raise NotImplementedError
    
    def begin_training(self, train_data):
        return self.Trainer(self, train_data)

    def pipe(self, stream, batch_size=1000):
        for batch in util.minibatch(stream, batch_size):
            ys = self.predict_batch(batch)
            for y in ys:
                yield y

    def update(self, stream, batch_size=1000):
        for X, y in util.minibatch(stream, batch_size=batch_size):
            output, finish_update = self.begin_update(X)
            gradient = finish_update(y)
            yield gradient
    
    def average_params(self, optimizer):
        pass

    def begin_update(self, X, dropout=0.0):
        raise NotImplementedError
    
    @property
    def is_initialized(self):
        return True


class Network(Model):
    '''A model that chains together other Models.'''
    name = 'mlp'

    @property
    def input_shape(self):
        return self.layers[0].input_shape

    @property
    def output_shape(self):
        return self.layers[-1].output_shape

    @property
    def nr_weight(self):
        return sum(layer.nr_weight for layer in self.layers)

    def setup(self, *layers, **kwargs):
        for i, layer in enumerate(layers):
            if isinstance(layer, Model):
                self.layers.append(layer)
            else:
                self.layers.append(layer(**kwargs))
        self.set_weights(initialize=True)
        self.set_gradient()

    def set_weights(self, example=None, data=None, initialize=True):
        if data is None:
            self.data = self.ops.allocate_pool(self.nr_weight,
                            name=(self.name, 'pool'))
        else:
            self.data = data
        for layer in self.layers:
            layer.set_weights(data=self.data, initialize=initialize)
            layer.data = None

    def set_gradient(self, data=None):
        if data is None:
            self.d_data = self.ops.allocate_pool(self.nr_weight,
                            name=(self.name, 'd_pool'))
        else:
            self.d_data = data
        for layer in self.layers:
            layer.set_gradient(data=self.d_data)
            layer.d_data = None

    def predict_batch(self, X):
        for layer in self.layers:
            X = layer.predict_batch(X)
        return X

    def begin_update(self, X, dropout=0.0):
        callbacks = []
        for layer in self.layers:
            X, finish_update = layer.begin_update(X, dropout=dropout)
            callbacks.append(finish_update)
        return X, self._get_finish_update(callbacks)

    def average_params(self, optimizer):
        for layer in self.layers:
            layer.average_params(optimizer)
        if self.data is not None and ('data', self.name) in optimizer.averages:
            self.data.data[:] = optimizer.averages[('data', self.name)]

    def _get_finish_update(self, callbacks):
        def finish_update(gradient, optimizer, **kwargs):
            for callback in reversed(callbacks):
                gradient = callback(gradient, optimizer=optimizer, **kwargs)
            if optimizer is not None and self.data is not None:
                optimizer(self.data.data, self.d_data.data,
                    key=('data', self.name), **kwargs)
            return gradient
        return finish_update



from collections import defaultdict
import numpy


def linear_decay(rate, decay, nr_upd):
    return rate * 1./(1. + decay * nr_upd)


def update_averages(averages, key, weights, nr_upd, max_decay=0.9999):
    decay = (1. + nr_upd) / (10. + nr_upd)
    decay = min(decay, max_decay)

    if key not in averages:
        averages[key] = numpy.zeros(weights.shape)
    avg = averages[key]
    avg -= (1-decay) * (avg - weights)


def clip_gradient(gradient, threshold):
    grad_norm = numpy.linalg.norm(gradient)
    if grad_norm >= threshold:
        gradient *= threshold / grad_norm


class SGD(object):
    def __init__(self, ops, lr, momentum=0.0, decay=0.0, **settings):
        self.ops = ops
        self.alpha = lr
        self.mu = momentum
        self.decay = decay
        self.momentums = {}
        self.averages = {}
        self.nr_update = defaultdict(int)

    def __call__(self, weights, gradient, key=None):
        self.nr_update[key] += 1
        nr_upd = self.nr_update[key]
        lr = self.lr(nr_upd)
        if key is None or self.mu == 0.0:
            weights -= lr * gradient
            gradient.fill(0)
        else:
            if key not in self.momentums:
                self.momentums[key] = numpy.zeros(weights.shape)
            momentum = self.momentums[key]
            momentum *= self.mu
            momentum += gradient * lr
            weights -= momentum
            gradient.fill(0)
        update_averages(self.averages, key, weights, nr_upd)

    def lr(self, nr_upd):
        return linear_decay(self.alpha, self.decay, nr_upd)


class Adam(object):
    def __init__(self, ops, lr, beta1=0.90, beta2=0.999, eps=1e-08, decay=0.0):
        self.ops = ops
        self.mom1 = {}
        self.mom2 = {}
        self.averages = {}
        self.nr_update = defaultdict(int)
        self.last_seen = defaultdict(int)
        self.alpha = lr
        self.b1 = beta1
        self.b2 = beta2
        self.eps = eps
        self.decay = decay
        self.d = 1.
        self.f = 0.

    def lr(self, nr_upd):
        alpha = linear_decay(self.alpha, self.decay, nr_upd)
        fix1 = 1.- (self.b1 ** nr_upd)
        fix2 = 1.- (self.b2 ** nr_upd)
        return alpha * numpy.sqrt(fix2) / fix1

    @property
    def nr_iter(self):
        if not self.nr_update:
            return 0
        return max(self.nr_update.values())

    def __call__(self, weights, gradient, key=None):
        assert key is not None
        assert len(gradient) >= 1
        assert not self.ops.xp.isnan(weights).any()
        assert not self.ops.xp.isnan(gradient).any()
        if key not in self.mom1:
            self.mom1[key] = self.ops.allocate(weights.shape)
        if key not in self.mom2:
            self.mom2[key] = self.ops.allocate(weights.shape)
        self.nr_update[key] += 1
        nr_upd = self.nr_update[key]

        clip_gradient(gradient, len(gradient) / 100.)
        mom1 = self.mom1[key]
        mom1 *= self.b1
        mom1 += (gradient * (1-self.b1))

        mom2 = self.mom2[key]
        mom2 *= self.b2
        mom2 += (1-self.b2) * gradient ** 2
        
        lr = self.lr(nr_upd)
        weights -= lr * mom1 / (self.d * numpy.sqrt(mom2) + self.eps)
        gradient.fill(0)
        update_averages(self.averages, key, weights, nr_upd)

    def set_loss(self, loss):
        pass


class Eve(Adam):
    def __init__(self, *args, **kwargs):
        Adam.__init__(self, *args, **kwargs)
        self.b3 = 0.999
        self.lower_threshold = 0.1
        self.upper_threshold = 10
        self.d = 1.
        self.f = None

    def set_loss(self, loss):
        if self.f is None:
            self.f = loss
            return
        old_f = self.f
        d = self.d
        c = self._get_c(loss, old_f)
        new_f = c * loss
        r = abs(new_f - old_f) / min(new_f, old_f)
        new_d = d + (1 - self.b3) * (r - d)
        self.d = new_d
        self.f = new_f

    def _get_c(self, loss, old_f):
        if loss < old_f:
            delta = self.lower_threshold + 1.
            Delta = self.upper_threshold + 1.
        else:
            delta = 1. / (self.upper_threshold + 1.)
            Delta = 1. / (self.lower_threshold + 1.)
        return min(max(delta, loss / old_f), Delta)



# https://raw.githubusercontent.com/fchollet/keras/master/keras/datasets/mnist.py
# Copyright Francois Chollet, Google, others (2015)
# Under MIT license

import gzip
from .keras_data_utils import get_file
import sys


def load_mnist(path='mnist.pkl.gz'):
    from six.moves import cPickle
    path = get_file(path,
        origin='https://s3.amazonaws.com/img-datasets/mnist.pkl.gz')

    if path.endswith('.gz'):
        f = gzip.open(path, 'rb')
    else:
        f = open(path, 'rb')

    if sys.version_info < (3,):
        data = cPickle.load(f)
    else:
        data = cPickle.load(f, encoding='bytes')

    f.close()
    return data  # (X_train, y_train), (X_test, y_test)



import pytest

from ...base import Model

class MockOps(object):
    pass


class Subclass(Model):
    @property
    def shape(self):
        return self._shape

    @property
    def nr_out(self):
        return self.shape[0]

    @property
    def nr_in(self):
        return self.shape[1]

    def setup(self, *shape, **kwargs):
        self._shape = tuple(shape)

    def predict_batch(self, X):
        return ['label' for _ in range(len(X))]

    def predict_one(self, x):
        return 'label'

    def check_shape(self, x, is_batch):
        pass

    def is_batch(self, x):
        return True if isinstance(x, list) else False


def test_specify_shape():
    sub = Subclass(2, 2, ops=MockOps())
    assert sub.shape == (2,2)
    assert sub.nr_out == 2
    assert sub.nr_in == 2


def test_predict():
    sub = Subclass(2, 2, ops=MockOps())
    assert sub('question') == 'label'
    assert sub(['q1', 'w2']) == ['label', 'label']

    for result in sub.pipe(['q1', 'w2']):
        assert result == 'label'


def test_begin_update():
    sub = Subclass(2, 2, ops=MockOps())
    with pytest.raises(NotImplementedError):
        sub.begin_update('question', dropout=0.0)
